import hashlib
import json
import uuid
from tempfile import NamedTemporaryFile

import pytest
import requests

from pulpcore.client.pulp_rpm import (
    AcsRpmApi,
    ContentAdvisoriesApi,
    ContentDistributionTreesApi,
    ContentPackagecategoriesApi,
    ContentPackagegroupsApi,
    ContentPackagelangpacksApi,
    ContentPackagesApi,
    ContentModulemdsApi,
    ContentModulemdDefaultsApi,
    ContentModulemdObsoletesApi,
    RemotesUlnApi,
    RpmCopyApi,
    RpmCompsApi,
    RpmRepositorySyncURL,
)

from pulp_rpm.tests.functional.constants import (
    BASE_TEST_JSON,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_SIGNED_URL,
    RPM_MODULAR_FIXTURE_URL,
    RPM_SIGNED_FIXTURE_URL,
)
from pulp_rpm.tests.functional.utils import init_signed_repo_configuration


@pytest.fixture(scope="session")
def rpm_ulnremote_api(rpm_client):
    """Fixture for RPM remote API."""
    return RemotesUlnApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_acs_api(rpm_client):
    """Fixture for RPM alternate content source API."""
    return AcsRpmApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_advisory_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentAdvisoriesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_category_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagecategoriesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_groups_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagegroupsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_package_lang_packs_api(rpm_client):
    """Fixture for RPM distribution API."""
    return ContentPackagelangpacksApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_modulemd_api(rpm_client):
    """Fixture for RPM Modulemd API."""
    return ContentModulemdsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_modulemd_defaults_api(rpm_client):
    """Fixture for RPM ModulemdDefault API."""
    return ContentModulemdDefaultsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_modulemd_obsoletes_api(rpm_client):
    """Fixture for RPM ModulemdObsolete API."""
    return ContentModulemdObsoletesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_comps_api(rpm_client):
    """Fixture for RPM Comps API."""
    return RpmCompsApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_content_distribution_trees_api(rpm_client):
    return ContentDistributionTreesApi(rpm_client)


@pytest.fixture(scope="session")
def rpm_copy_api(rpm_client):
    return RpmCopyApi(rpm_client)


@pytest.fixture
def signed_artifact(pulpcore_bindings, tmp_path):
    data = requests.get(RPM_SIGNED_URL).content
    artifacts = pulpcore_bindings.ArtifactsApi.list(
        sha256=hashlib.sha256(data).hexdigest(), limit=1
    )
    try:
        return artifacts.results[0].to_dict()
    except IndexError:
        pass

    temp_file = tmp_path / str(uuid.uuid4())
    temp_file.write_bytes(data)
    return pulpcore_bindings.ArtifactsApi.create(temp_file).to_dict()


@pytest.fixture
def rpm_artifact_factory(pulpcore_bindings, gen_object_with_cleanup, pulp_domain_enabled, tmp_path):
    """Return an artifact created from uploading an RPM file."""

    def _rpm_artifact_factory(url=RPM_SIGNED_URL, pulp_domain=None):
        temp_file = tmp_path / str(uuid.uuid4())
        temp_file.write_bytes(requests.get(url).content)
        kwargs = {}
        if pulp_domain:
            if not pulp_domain_enabled:
                raise RuntimeError("Server does not have domains enabled.")
            kwargs["pulp_domain"] = pulp_domain
        return gen_object_with_cleanup(pulpcore_bindings.ArtifactsApi, temp_file, **kwargs)

    return _rpm_artifact_factory


@pytest.fixture
def rpm_package_factory(
    gen_object_with_cleanup,
    pulp_domain_enabled,
    rpm_package_api,
):
    """Return a Package created from uploading an RPM file."""

    def _rpm_package_factory(url=RPM_SIGNED_URL, pulp_domain=None):
        with NamedTemporaryFile() as file_to_upload:
            file_to_upload.write(requests.get(url).content)
            file_to_upload.flush()
            upload_attrs = {"file": file_to_upload.name}

            kwargs = {}
            if pulp_domain:
                if not pulp_domain_enabled:
                    raise RuntimeError("Server does not have domains enabled.")
                kwargs["pulp_domain"] = pulp_domain

            return gen_object_with_cleanup(rpm_package_api, **upload_attrs, **kwargs)

    return _rpm_package_factory


@pytest.fixture(scope="class")
def rpm_unsigned_repo_immediate(init_and_sync):
    repo, _ = init_and_sync()
    return repo


@pytest.fixture(scope="class")
def rpm_unsigned_repo_on_demand(init_and_sync):
    repo, _ = init_and_sync(policy="on_demand")
    return repo


@pytest.fixture(scope="class")
def rpm_modular_repo_on_demand(init_and_sync):
    repo, _ = init_and_sync(url=RPM_MODULAR_FIXTURE_URL, policy="on_demand")
    return repo


@pytest.fixture(scope="class")
def rpm_kickstart_repo_immediate(init_and_sync):
    repo, _ = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL)
    return repo


@pytest.fixture(scope="session")
def rpm_metadata_signing_service(pulpcore_bindings):
    results = pulpcore_bindings.SigningServicesApi.list(name="sign-metadata")
    signing_service = None
    if results.count == 0:
        result = init_signed_repo_configuration()
        if result.returncode == 0:
            results = pulpcore_bindings.SigningServicesApi.list(name="sign-metadata")
    if results.count == 1:
        signing_service = results.results[0]

    return signing_service


@pytest.fixture
def upload_wrong_file_type(rpm_advisory_api):
    def _upload(remote_path):
        with NamedTemporaryFile() as file_to_upload:
            file_to_upload.write(requests.get(remote_path).content)
            file_to_upload.flush()
            upload_attrs = {"file": file_to_upload.name}
            return rpm_advisory_api.create(**upload_attrs)

    return _upload


@pytest.fixture
def upload_advisory_factory(
    add_to_cleanup,
    monitor_task,
    pulp_domain_enabled,
    rpm_advisory_api,
):
    """Upload advisory from a json file, return advisory, vers_href, and id-used."""

    def _upload_advisory_factory(
        advisory=BASE_TEST_JSON, repository=None, pulp_domain=None, set_id=False, use_id=None
    ):
        kwargs = {}
        if pulp_domain:
            if not pulp_domain_enabled:
                raise RuntimeError("Server does not have domains enabled.")
            kwargs["pulp_domain"] = pulp_domain

        with NamedTemporaryFile("w+") as file_to_upload:
            json_advisory = json.loads(advisory)
            if set_id or use_id:
                json_advisory["id"] = use_id if use_id else str(uuid.uuid4())
            used_id = json_advisory["id"]
            json.dump(json_advisory, file_to_upload)
            file_to_upload.flush()
            upload_attrs = {"file": file_to_upload.name}
            if repository:
                upload_attrs["repository"] = repository.pulp_href
            file_to_upload.flush()
            response = rpm_advisory_api.create(**upload_attrs, **kwargs)

        task_rslt = monitor_task(response.task)
        if repository:
            assert 2 == len(task_rslt.created_resources)
        else:
            assert 1 == len(task_rslt.created_resources)

        vers_href = None
        advisory_href = None
        for rsrc in task_rslt.created_resources:
            if "versions" in rsrc:
                vers_href = rsrc
            elif "advisories" in rsrc:
                advisory_href = rsrc
        assert advisory_href

        add_to_cleanup(rpm_advisory_api, advisory_href)
        entity = rpm_advisory_api.read(advisory_href)
        return entity, vers_href, used_id

    return _upload_advisory_factory


@pytest.fixture
def assert_uploaded_advisory(rpm_advisory_api):
    """List advisories for a given version-href, and assert that the specified ID is therein."""

    def _from_results(advisory_id, vers_href):
        advisories = rpm_advisory_api.list(id=advisory_id, repository_version=vers_href)
        assert 1 == len(advisories.results)
        return advisories.results[0].pulp_href, vers_href

    return _from_results


@pytest.fixture
def setup_domain(
    gen_object_with_cleanup, pulpcore_bindings, rpm_rpmremote_api, rpm_repository_api, monitor_task
):
    def _setup_domain(sync=True, url=RPM_SIGNED_FIXTURE_URL, pulp_domain=None):
        if not pulp_domain:
            body = {
                "name": str(uuid.uuid4()),
                "storage_class": "pulpcore.app.models.storage.FileSystem",
                "storage_settings": {"MEDIA_ROOT": "/var/lib/pulp/media/"},
            }
            pulp_domain = gen_object_with_cleanup(pulpcore_bindings.DomainsApi, body)

        remote = gen_object_with_cleanup(
            rpm_rpmremote_api, {"name": str(uuid.uuid4()), "url": url}, pulp_domain=pulp_domain.name
        )
        src = gen_object_with_cleanup(
            rpm_repository_api,
            {"name": str(uuid.uuid4()), "remote": remote.pulp_href},
            pulp_domain=pulp_domain.name,
        )

        if sync:
            sync_url = RpmRepositorySyncURL()
            monitor_task(rpm_repository_api.sync(src.pulp_href, sync_url).task)
            src = rpm_repository_api.read(src.pulp_href)

        dest = gen_object_with_cleanup(
            rpm_repository_api, {"name": str(uuid.uuid4())}, pulp_domain=pulp_domain.name
        )
        return pulp_domain, remote, src, dest

    return _setup_domain


@pytest.fixture
def cleanup_domains(pulpcore_bindings, monitor_task, rpm_repository_api):
    def _cleanup_domains(
        domains,
        content_api_client=None,
        cleanup_repositories=False,
        repository_api_client=rpm_repository_api,
    ):
        for domain in domains:
            # clean up each domain specified
            if domain:
                if cleanup_repositories:
                    # Delete repos from the domain
                    for repo in repository_api_client.list(pulp_domain=domain.name).results:
                        monitor_task(repository_api_client.delete(repo.pulp_href).task)
                # Let orphan-cleanup reap the resulting abandoned content
                monitor_task(
                    pulpcore_bindings.OrphansCleanupApi.cleanup(
                        {"orphan_protection_time": 0}, pulp_domain=domain.name
                    ).task
                )

        if content_api_client:
            # IF we have a client, check that each domain is empty of that kind-of entity
            for domain in domains:
                if domain:
                    assert content_api_client.list(pulp_domain=domain.name).count == 0

    return _cleanup_domains



# Generated by Django 3.2.13 on 2022-05-02 07:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0041_modulemdobsolete'),
    ]

    operations = [
        migrations.AlterField(
            model_name='repometadatafile',
            name='data_type',
            field=models.TextField(),
        ),
    ]



"""Tests that publish rpm plugin repositories."""

from random import choice


from pulp_smash import config
from pulp_smash.pulp3.bindings import PulpTestCase, monitor_task
from pulp_smash.pulp3.utils import (
    gen_repo,
    get_content,
    gen_distribution,
    get_versions,
    modify_repo,
)

from pulp_rpm.tests.functional.constants import RPM_PACKAGE_CONTENT_NAME
from pulp_rpm.tests.functional.utils import gen_rpm_client, gen_rpm_remote
from pulp_rpm.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_rpm import (
    DistributionsRpmApi,
    PublicationsRpmApi,
    RepositoriesRpmApi,
    RpmRepositorySyncURL,
    RemotesRpmApi,
    RpmRpmPublication,
)
from pulpcore.client.pulp_rpm.exceptions import ApiException


class PublishAnyRepoVersionTestCase(PulpTestCase):
    """Test whether a particular repository version can be published.

    This test targets the following issues:

    * `Pulp #3324 <https://pulp.plan.io/issues/3324>`_
    * `Pulp Smash #897 <https://github.com/pulp/pulp-smash/issues/897>`_
    """

    def test_all(self):
        """Test whether a particular repository version can be published.

        1. Create a repository with at least 2 repository versions.
        2. Create a publication by supplying the latest ``repository_version``.
        3. Assert that the publication ``repository_version`` attribute points
           to the latest repository version.
        4. Create a publication by supplying the non-latest ``repository_version``.
        5. Assert that the publication ``repository_version`` attribute points
           to the supplied repository version.
        6. Assert that an exception is raised when providing two different
           repository versions to be published at same time.
        """
        cfg = config.get_config()
        client = gen_rpm_client()
        repo_api = RepositoriesRpmApi(client)
        remote_api = RemotesRpmApi(client)
        publications = PublicationsRpmApi(client)
        distributions = DistributionsRpmApi(client)

        body = gen_rpm_remote()
        remote = remote_api.create(body)

        repo = repo_api.create(gen_repo())

        repository_sync_data = RpmRepositorySyncURL(remote=remote.pulp_href)
        sync_response = repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

        # Step 1
        repo = repo_api.read(repo.pulp_href)
        repo_content = get_content(repo.to_dict())[RPM_PACKAGE_CONTENT_NAME][:-1]
        for rpm_content in repo_content:
            modify_repo(cfg, repo.to_dict(), remove_units=[rpm_content])
        version_hrefs = tuple(ver["pulp_href"] for ver in get_versions(repo.to_dict()))
        non_latest = choice(version_hrefs[1:-1])

        # Step 2
        publish_data = RpmRpmPublication(repository=repo.pulp_href)
        publish_response = publications.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]
        publication = publications.read(publication_href)

        # Step 3
        self.assertEqual(publication.repository_version, version_hrefs[-1])

        # Step 4
        publish_data.repository_version = non_latest
        publish_data.repository = None
        publish_response = publications.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]
        publication = publications.read(publication_href)

        # Step 5
        body = gen_distribution()
        body["base_path"] = "pulp_pre_upgrade_test"
        body["publication"] = publication.pulp_href

        distribution_response = distributions.create(body)
        created_resources = monitor_task(distribution_response.task).created_resources
        distribution = distributions.read(created_resources[0])

        # Step 6
        self.assertEqual(publication.repository_version, non_latest)

        # Step 7
        with self.assertRaises(ApiException):
            body = {"repository": repo.pulp_href, "repository_version": non_latest}
            publications.create(body)

        # Step 8
        url = cfg.get_content_host_base_url() + "/pulp/content/pulp_pre_upgrade_test/"
        self.assertEqual(url, distribution.base_url, url)



"""Tests that publish rpm plugin repositories."""

import os
import pytest

import collections
from lxml import etree
from django.conf import settings
from xml.etree import ElementTree
import requests

import xmltodict
import dictdiffer

from pulp_smash.pulp3.utils import gen_repo, gen_distribution

from pulp_rpm.tests.functional.constants import (
    RPM_ALT_LAYOUT_FIXTURE_URL,
    RPM_COMPLEX_FIXTURE_URL,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_KICKSTART_REPOSITORY_ROOT_CONTENT,
    RPM_REPO_METADATA_FIXTURE_URL,
    RPM_LONG_UPDATEINFO_FIXTURE_URL,
    RPM_MODULAR_FIXTURE_URL,
    RPM_NAMESPACES,
    RPM_REFERENCES_UPDATEINFO_URL,
    RPM_RICH_WEAK_FIXTURE_URL,
    RPM_SHA512_FIXTURE_URL,
    RPM_UNSIGNED_FIXTURE_URL,
    SRPM_UNSIGNED_FIXTURE_URL,
)
from pulp_rpm.tests.functional.utils import gen_rpm_remote, download_and_decompress_file
from pulp_rpm.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_rpm import RpmRepositorySyncURL, RpmRpmPublication
from pulpcore.client.pulp_rpm.exceptions import ApiException


class TestPublishWithUnsignedRepoSyncedImmediate:
    @pytest.mark.parallel
    def test_publish_any_repo_version(
        self,
        rpm_unsigned_repo_immediate,
        rpm_publication_api,
        monitor_task,
    ):
        """Test whether a particular repository version can be published."""
        publish_data = RpmRpmPublication(repository=rpm_unsigned_repo_immediate.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        assert (
            rpm_publication_api.read(publication_href).repository_version
            == rpm_unsigned_repo_immediate.latest_version_href
        )

        non_latest = f"{rpm_unsigned_repo_immediate.pulp_href}versions/0/"

        publish_data.repository_version = non_latest
        publish_data.repository = None
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]
        publication = rpm_publication_api.read(publication_href)

        assert publication.repository_version == non_latest

        with pytest.raises(ApiException):
            body = {
                "repository": rpm_unsigned_repo_immediate.pulp_href,
                "repository_version": non_latest,
            }
            rpm_publication_api.create(body)

    @pytest.mark.parametrize("compression_type,compression_ext", (("gz", ".gz"), ("zstd", ".zst")))
    @pytest.mark.parallel
    def test_publish_with_compression_types(
        self,
        compression_type,
        compression_ext,
        rpm_unsigned_repo_immediate,
        rpm_publication_api,
        gen_object_with_cleanup,
        rpm_distribution_api,
        monitor_task,
    ):
        """Sync and publish an RPM repository w/ zstd compression and verify it exists."""
        # 1. Publish and distribute
        publish_data = RpmRpmPublication(
            repository=rpm_unsigned_repo_immediate.pulp_href, compression_type=compression_type
        )
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        # 2. Check "primary", "filelists", "other", "updateinfo" have correct compression ext
        for md_type, md_href in self.get_repomd_metadata_urls(distribution.base_url).items():
            if md_type in ("primary", "filelists", "other", "updateinfo"):
                assert md_href.endswith(compression_ext)

    @pytest.mark.parallel
    def test_validate_no_checksum_tag(
        self,
        rpm_unsigned_repo_immediate,
        rpm_publication_api,
        gen_object_with_cleanup,
        rpm_distribution_api,
        monitor_task,
    ):
        """Sync and publish an RPM repository and verify the checksum.

        TODO: replace with https://github.com/pulp/pulp_rpm/issues/2957
        """
        # 1. Publish and distribute
        publish_data = RpmRpmPublication(repository=rpm_unsigned_repo_immediate.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        # 2. check the tag 'sum' is not present in updateinfo.xml
        update_xml_url = self.get_repomd_metadata_urls(distribution.base_url)["updateinfo"]
        update_xml = download_and_decompress_file(
            os.path.join(distribution.base_url, update_xml_url)
        )
        update_info_content = ElementTree.fromstring(update_xml)

        tags = {elem.tag for elem in update_info_content.iter()}
        assert "sum" not in tags, update_info_content

    @staticmethod
    def get_repomd_metadata_urls(repomd_url: str):
        """
        Helper function to get hrefs of repomd types.

        Example:
            ```
            >>> get_repomd_metadata_urls(distribution.base_url)
            {
                "primary": "repodata/.../primary.xml.gz",
                "filelists": "repodata/.../filelists.xml.gz",
                ...
            }
            ```
        """
        # XML Reference:
        # <ns0:repomd xmlns:ns0="http://linux.duke.edu/metadata/repo">
        #     <ns0:data type="primary">
        #         <ns0:checksum type="sha256">[…]</ns0:checksum>
        #         <ns0:location href="repodata/[…]-primary.xml.gz" />
        #         …
        #     </ns0:data>
        #     …
        repomd_xml = requests.get(os.path.join(repomd_url, "repodata/repomd.xml")).text
        repomd = ElementTree.fromstring(repomd_xml)
        xpath_data = "{{{}}}data".format(RPM_NAMESPACES["metadata/repo"])
        xpath_location = "{{{}}}location".format(RPM_NAMESPACES["metadata/repo"])
        hrefs = {}
        for elem in repomd.findall(xpath_data):
            md_type = elem.get("type")
            hrefs[md_type] = elem.find(xpath_location).get("href")
        return hrefs


@pytest.fixture(scope="class")
def assert_created_publication(init_and_sync, rpm_publication_api, monitor_task):
    def _assert_created_publication(url):
        """Sync and publish an RPM repository given a feed URL."""
        repo, _ = init_and_sync(url=url)
        publish_data = RpmRpmPublication(repository=repo.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        assert monitor_task(publish_response.task).created_resources != []

    return _assert_created_publication


@pytest.mark.parallel
def test_publish_rpm_rich_weak(assert_created_publication):
    """Sync and publish an RPM rich weak repository."""
    assert_created_publication(RPM_RICH_WEAK_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_rpm_long_updateinfo(assert_created_publication):
    """Sync and publish an RPM long updateinfo repository."""
    assert_created_publication(RPM_LONG_UPDATEINFO_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_rpm_custom_metadata(assert_created_publication):
    """Sync and publish an RPM with custom metadata files."""
    assert_created_publication(RPM_REPO_METADATA_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_rpm_alt_layout(assert_created_publication):
    """Sync and publish an RPM alternate layout repository."""
    assert_created_publication(RPM_ALT_LAYOUT_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_rpm_sha512(assert_created_publication):
    """Sync and publish an RPM SHA256 repository."""
    assert_created_publication(RPM_SHA512_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_srpm(assert_created_publication):
    """Sync and publish an SRPM repository."""
    assert_created_publication(SRPM_UNSIGNED_FIXTURE_URL)


@pytest.mark.parallel
def test_publish_references_update(assert_created_publication):
    """Sync/publish a repo where ``updateinfo.xml`` contains references."""
    # TODO: check the updateinfo metadata to confirm the references are there.
    assert_created_publication(RPM_REFERENCES_UPDATEINFO_URL)


@pytest.mark.parametrize("repo_url", [RPM_COMPLEX_FIXTURE_URL, RPM_MODULAR_FIXTURE_URL])
def test_complex_repo_core_metadata(
    repo_url,
    init_and_sync,
    rpm_publication_api,
    gen_object_with_cleanup,
    rpm_distribution_api,
    monitor_task,
    delete_orphans_pre,
):
    """Test the "complex" fixture that covers more of the metadata cases.

    The standard fixtures have no changelogs and don't cover "ghost" files. The repo
    with the "complex-package" does, and also does a better job of covering rich deps
    and other atypical metadata.
    """

    # create repo and remote
    repo, _ = init_and_sync(url=repo_url, policy="on_demand")

    # publish
    publish_data = RpmRpmPublication(repository=repo.pulp_href)
    publish_response = rpm_publication_api.create(publish_data)
    created_resources = monitor_task(publish_response.task).created_resources
    publication_href = created_resources[0]

    # distribute
    body = gen_distribution(publication=publication_href)
    distribution = gen_object_with_cleanup(rpm_distribution_api, body)

    # Download and parse the metadata.
    original_repomd = ElementTree.fromstring(
        requests.get(os.path.join(repo_url, "repodata/repomd.xml")).text
    )

    reproduced_repomd = ElementTree.fromstring(
        requests.get(os.path.join(distribution.base_url, "repodata/repomd.xml")).text
    )

    def get_metadata_content(base_url, repomd_elem, meta_type):
        """Return the text contents of metadata file.

        Provided a url, a repomd root element, and a metadata type, locate the metadata
        file's location href, download it from the provided url, un-gzip it, parse it, and
        return the root element node.

        Don't use this with large repos because it will blow up.
        """
        # <ns0:repomd xmlns:ns0="http://linux.duke.edu/metadata/repo">
        #     <ns0:data type="primary">
        #         <ns0:checksum type="sha256">[…]</ns0:checksum>
        #         <ns0:location href="repodata/[…]-primary.xml.gz" />
        #         …
        #     </ns0:data>
        #     …
        xpath = "{{{}}}data".format(RPM_NAMESPACES["metadata/repo"])
        data_elems = [elem for elem in repomd_elem.findall(xpath) if elem.get("type") == meta_type]
        if not data_elems:
            return None

        xpath = "{{{}}}location".format(RPM_NAMESPACES["metadata/repo"])
        location_href = data_elems[0].find(xpath).get("href")

        return download_and_decompress_file(os.path.join(base_url, location_href))

    # Convert the metadata into a more workable form and then compare.
    for metadata_file in ["primary", "filelists", "other"]:
        original_metadata = get_metadata_content(repo_url, original_repomd, metadata_file)
        generated_metadata = get_metadata_content(
            distribution.base_url, reproduced_repomd, metadata_file
        )

        _compare_xml_metadata_file(original_metadata, generated_metadata, metadata_file)

    # =================

    original_modulemds = get_metadata_content(repo_url, original_repomd, "modules")
    generated_modulemds = get_metadata_content(distribution.base_url, reproduced_repomd, "modules")

    assert bool(original_modulemds) == bool(generated_modulemds)

    if original_modulemds:
        # compare list of modulemd, modulemd-defaults, and modulemd-obsoletes after sorting them
        original_modulemds = sorted(original_modulemds.decode().split("---")[1:])
        generated_modulemds = sorted(generated_modulemds.decode().split("---")[1:])

        assert original_modulemds == generated_modulemds

    # ===================

    # TODO: make this deeper
    original_updateinfo = get_metadata_content(repo_url, original_repomd, "updateinfo")
    generated_updateinfo = get_metadata_content(
        distribution.base_url, reproduced_repomd, "updateinfo"
    )
    assert bool(original_updateinfo) == bool(generated_updateinfo)


def _compare_xml_metadata_file(original_metadata_text, generated_metadata_text, meta_type):
    """Compare two metadata files.

    First convert the metadata into a canonical form. We convert from XML to JSON, and then to
    a dict, and then apply transformations to the dict to standardize the ordering. Then, we
    perform comparisons and observe any differences.
    """
    metadata_block_names = {
        "primary": "metadata",
        "filelists": "filelists",
        "other": "otherdata",
    }
    subsection = metadata_block_names[meta_type]

    # First extract the package entries
    original_metadata = xmltodict.parse(
        original_metadata_text, dict_constructor=collections.OrderedDict
    )[subsection]["package"]
    generated_metadata = xmltodict.parse(
        generated_metadata_text, dict_constructor=collections.OrderedDict
    )[subsection]["package"]

    if not isinstance(original_metadata, list):
        original_metadata = [original_metadata]
    if not isinstance(generated_metadata, list):
        generated_metadata = [generated_metadata]

    # The other transformations are inside the package nodes - they differ by type of metadata
    if meta_type == "primary":
        # location_href gets rewritten by Pulp so we should ignore these differences
        ignore = {"location.@href", "format.file", "time.@file"}
        # we need to make sure all of the requirements are in the same order
        requirement_types = [
            "rpm:suggests",
            "rpm:recommends",
            "rpm:enhances",
            "rpm:provides",
            "rpm:requires",
            "rpm:obsoletes",
            "rpm:conflicts",
            "rpm:supplements",
        ]

        original_metadata = sorted(original_metadata, key=lambda x: x["checksum"]["#text"])
        generated_metadata = sorted(generated_metadata, key=lambda x: x["checksum"]["#text"])
        for md_file in [original_metadata, generated_metadata]:
            for pkg in md_file:
                for req in requirement_types:
                    if pkg["format"].get(req):
                        pkg["format"][req] = sorted(pkg["format"][req])
    elif meta_type == "filelists":
        ignore = {}
        original_metadata = sorted(original_metadata, key=lambda x: x["@pkgid"])
        generated_metadata = sorted(generated_metadata, key=lambda x: x["@pkgid"])
        # make sure the files are all in the same order and type and sort them
        # nodes with a "type" attribute in the XML become OrderedDicts, so we have to convert
        # them to a string representation.
        for md_file in [original_metadata, generated_metadata]:
            for pkg in md_file:
                if pkg.get("file"):
                    files = []
                    for f in pkg["file"]:
                        if isinstance(f, collections.OrderedDict):
                            files.append(
                                "{path} type={type}".format(path=f["@type"], type=f["#text"])
                            )
                        else:
                            files.append(f)

                    pkg["file"] = sorted(files)
    elif meta_type == "other":
        ignore = {}
        original_metadata = sorted(original_metadata, key=lambda x: x["@pkgid"])
        generated_metadata = sorted(generated_metadata, key=lambda x: x["@pkgid"])

        for md_file in [original_metadata, generated_metadata]:
            for pkg in md_file:
                # make sure the changelogs are in the same order
                if pkg.get("changelog"):
                    pkg["changelog"] = sorted(pkg["changelog"], key=lambda x: x["@author"])

    # The metadata dicts should now be consistently ordered. Check for differences.
    for original, generated in zip(original_metadata, generated_metadata):
        diff = dictdiffer.diff(original, generated, ignore=ignore)
        assert list(diff) == [], list(diff)


@pytest.mark.parallel
@pytest.mark.parametrize("mirror", [True, False], ids=["mirror", "standard"])
def test_distribution_tree_metadata_publish(
    mirror,
    gen_object_with_cleanup,
    rpm_repository_api,
    rpm_rpmremote_api,
    rpm_distribution_api,
    monitor_task,
):
    """Test the "complex" fixture that covers more of the metadata cases.

    The standard fixtures have no changelogs and don't cover "ghost" files. The repo
    with the "complex-package" does, and also does a better job of covering rich deps
    and other atypical metadata.
    """
    from configparser import ConfigParser

    # 1. create repo and remote
    repo = gen_object_with_cleanup(rpm_repository_api, gen_repo(autopublish=not mirror))

    body = gen_rpm_remote(RPM_KICKSTART_FIXTURE_URL, policy="on_demand")
    remote = gen_object_with_cleanup(rpm_rpmremote_api, body)

    # 2, 3. Sync and publish
    repository_sync_data = RpmRepositorySyncURL(remote=remote.pulp_href, mirror=mirror)

    sync_response = rpm_repository_api.sync(repo.pulp_href, repository_sync_data)
    created_resources = monitor_task(sync_response.task).created_resources

    publication_href = [r for r in created_resources if "publication" in r][0]

    body = gen_distribution(publication=publication_href)
    distribution = gen_object_with_cleanup(rpm_distribution_api, body)

    # 4. Download and parse the metadata.
    original_treeinfo = requests.get(os.path.join(RPM_KICKSTART_FIXTURE_URL, ".treeinfo")).text
    generated_treeinfo = requests.get(os.path.join(distribution.base_url, ".treeinfo")).text

    config = ConfigParser()
    config.optionxform = str  # by default it will cast keys to lower case
    config.read_string(original_treeinfo)
    original_treeinfo = config._sections

    config = ConfigParser()
    config.optionxform = str  # by default it will cast keys to lower case
    config.read_string(generated_treeinfo)
    generated_treeinfo = config._sections

    # 5, 6. Re-arrange the metadata so that it can be compared, and do the comparison.
    # TODO: These really should be in the same order they were in originally.
    # https://pulp.plan.io/issues/9208
    for metadata_dict in [original_treeinfo, generated_treeinfo]:
        metadata_dict["general"]["variants"] = ",".join(
            sorted(metadata_dict["general"]["variants"].split(","))
        )
        metadata_dict["tree"]["variants"] = ",".join(
            sorted(metadata_dict["tree"]["variants"].split(","))
        )

    diff = dictdiffer.diff(original_treeinfo, generated_treeinfo)
    differences = []

    # skip any differences that are "correct" i.e. rewritten "repository" and "packages" paths
    for d in diff:
        (diff_type, diff_name, _, new_value) = (d[0], d[1], d[2][0], d[2][1])
        # ('change', 'variant-Land.packages', ('Packages', 'Land/Packages'))
        if diff_type == "change":
            if diff_name.endswith(".packages") or diff_name.endswith(".repository"):
                # TODO: this is ignoring problems with the generated metadata
                # https://pulp.plan.io/issues/9208
                if "../" not in new_value:
                    continue

        differences.append(d)

    assert differences == [], differences

    # 7. Try downloading the files listed in the .treeinfo metadata, make sure they're
    # actually there.
    for path, checksum in original_treeinfo["checksums"].items():
        if path.startswith("fixtures"):
            # TODO: the .treeinfo metadata is actually wrong for these files, so we can't
            # check them because they won't be there.
            continue

        checksum_type, checksum = checksum.split(":")
        assert requests.get(os.path.join(distribution.base_url, path)).status_code == 200


@pytest.fixture
def get_checksum_types(
    init_and_sync,
    rpm_publication_api,
    gen_object_with_cleanup,
    rpm_distribution_api,
    monitor_task,
):
    """Sync and publish an RPM repository."""

    def _get_checksum_types(**kwargs):
        fixture_url = kwargs.get("fixture_url", RPM_UNSIGNED_FIXTURE_URL)
        package_checksum_type = kwargs.get("package_checksum_type")
        metadata_checksum_type = kwargs.get("metadata_checksum_type")
        policy = kwargs.get("policy", "immediate")

        # 1. create repo and remote
        repo, _ = init_and_sync(policy=policy, url=fixture_url, sync_policy="additive")

        # 2. Publish and distribute
        publish_data = RpmRpmPublication(
            repository=repo.pulp_href,
            package_checksum_type=package_checksum_type,
            metadata_checksum_type=metadata_checksum_type,
        )
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        repomd = ElementTree.fromstring(
            requests.get(os.path.join(distribution.base_url, "repodata/repomd.xml")).text
        )

        data_xpath = "{{{}}}data".format(RPM_NAMESPACES["metadata/repo"])
        data_elems = [elem for elem in repomd.findall(data_xpath)]

        repomd_checksum_types = {}
        primary_checksum_types = {}
        checksum_xpath = "{{{}}}checksum".format(RPM_NAMESPACES["metadata/repo"])
        for data_elem in data_elems:
            checksum_type = data_elem.find(checksum_xpath).get("type")
            repomd_checksum_types[data_elem.get("type")] = checksum_type
            if data_elem.get("type") == "primary":
                location_xpath = "{{{}}}location".format(RPM_NAMESPACES["metadata/repo"])
                primary_href = data_elem.find(location_xpath).get("href")
                primary = ElementTree.fromstring(
                    download_and_decompress_file(os.path.join(distribution.base_url, primary_href))
                )
                package_checksum_xpath = "{{{}}}checksum".format(RPM_NAMESPACES["metadata/common"])
                package_xpath = "{{{}}}package".format(RPM_NAMESPACES["metadata/common"])
                package_elems = [elem for elem in primary.findall(package_xpath)]
                pkg_checksum_type = package_elems[0].find(package_checksum_xpath).get("type")
                primary_checksum_types[package_elems[0].get("type")] = pkg_checksum_type

        return repomd_checksum_types, primary_checksum_types

    return _get_checksum_types


@pytest.mark.parallel
def test_publish_with_disallowed_checksum_type(rpm_unsigned_repo_on_demand, rpm_publication_api):
    """
    Sync and try to publish an RPM repository.

    - Sync repository with on_demand policy
    - Try to publish with 'sha384' checksum type
    - Publish should fail because 'sha384' is not allowed

    This test require disallowed 'sha384' checksum type from ALLOWED_CONTENT_CHECKSUMS settings.
    """
    if "sha384" in settings.ALLOWED_CONTENT_CHECKSUMS:
        pytest.skip(
            reason="Cannot check for the expected error if the 'sha384' checksum is allowed."
        )

    publish_data = RpmRpmPublication(
        repository=rpm_unsigned_repo_on_demand.pulp_href, package_checksum_type="sha384"
    )
    with pytest.raises(ApiException) as ctx:
        rpm_publication_api.create(publish_data)

    assert "Checksum must be one of the allowed checksum types" in ctx.value.body


@pytest.mark.parallel
def test_publish_with_unsupported_checksum_type(rpm_unsigned_repo_on_demand, rpm_publication_api):
    """
    Sync and try to publish an RPM repository.

    - Sync repository with on_demand policy
    - Try to publish with 'sha1' checksum type
    - Publish should fail because 'sha1' is not allowed
      (even though it is in ALLOWED_CONTENT_CHECKSUMS)
    """
    publish_data = RpmRpmPublication(
        repository=rpm_unsigned_repo_on_demand.pulp_href, package_checksum_type="sha1"
    )
    with pytest.raises(ApiException) as ctx:
        rpm_publication_api.create(publish_data)

    assert "Checksum must be one of the allowed checksum types" in ctx.value.body


@pytest.mark.parallel
def test_on_demand_unspecified_checksum_types(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        fixture_url=RPM_SHA512_FIXTURE_URL, policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        # hack to account for the fact that our md5 and sha512 repos use md5 and sha256
        # checksums for all metadata *except* updateinfo :(
        if not repomd_type == "updateinfo":
            assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        # since none of the packages in question have sha512 checksums, the
        # checksums they do have will be used instead. In this case, sha512.
        assert package_checksum_type == "sha512"


@pytest.mark.parallel
def test_immediate_unspecified_checksum_types(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        fixture_url=RPM_SHA512_FIXTURE_URL, policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        # hack to account for the fact that our md5 and sha512 repos use md5 and sha256
        # checksums for all metadata *except* updateinfo :(
        if not repomd_type == "updateinfo":
            assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha512"


def test_on_demand_specified_package_checksum_type(get_checksum_types, delete_orphans_pre):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha512", policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha256"

    for package, package_checksum_type in primary_checksum_types.items():
        # since none of the packages in question have sha512 checksums, the
        # checksums they do have will be used instead. In this case, sha256.
        assert package_checksum_type == "sha256"


@pytest.mark.parallel
def test_on_demand_specified_metadata_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        metadata_checksum_type="sha512", policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha256"


def test_on_demand_specified_metadata_and_package_checksum_type(
    get_checksum_types, delete_orphans_pre
):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha512", metadata_checksum_type="sha512", policy="on_demand"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        # since none of the packages in question have sha512 checksums, the
        # checksums they do have will be used instead. In this case, sha256.
        assert package_checksum_type == "sha256"


@pytest.mark.parallel
def test_immediate_specified_package_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha512", policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha256"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha512"


@pytest.mark.parallel
def test_immediate_specified_metadata_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        metadata_checksum_type="sha512", policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha256"


@pytest.mark.parallel
def test_immediate_specified_metadata_and_package_checksum_type(get_checksum_types):
    """Sync and publish an RPM repository and verify the checksum types."""
    repomd_checksum_types, primary_checksum_types = get_checksum_types(
        package_checksum_type="sha512", metadata_checksum_type="sha512", policy="immediate"
    )

    for repomd_type, repomd_checksum_type in repomd_checksum_types.items():
        assert repomd_checksum_type == "sha512"

    for package, package_checksum_type in primary_checksum_types.items():
        assert package_checksum_type == "sha512"


@pytest.mark.parallel
def test_directory_layout_distribute_with_modules(generate_distribution):
    """Ensure no more files or folders are present when distribute repository with modules."""
    distribution = generate_distribution(RPM_MODULAR_FIXTURE_URL)
    parser = etree.XMLParser(recover=True)
    repository = etree.fromstring(requests.get(distribution).text, parser=parser)
    # Get links from repository HTML
    # Each link is an item (file or directory) in repository root
    repository_root_items = []
    for elem in repository.iter():
        if elem.tag == "a" and not elem.text.startswith(".."):  # skip parent-dir if present
            repository_root_items.append(elem.attrib["href"])

    # Check if 'Packages' and 'repodata' are present
    # Trailing '/' is present for easier check
    assert "Packages/" in repository_root_items
    assert "repodata/" in repository_root_items
    # Only these three items should be present
    assert len(repository_root_items) == 2


@pytest.mark.parallel
def test_directory_layout_distribute_with_treeinfo(generate_distribution):
    """Ensure no more files or folders are present when distribute repository with treeinfo."""
    distribution = generate_distribution(RPM_KICKSTART_FIXTURE_URL)
    parser = etree.XMLParser(recover=True)
    repository = etree.fromstring(requests.get(distribution).text, parser=parser)
    # Get links from repository HTML
    # Each link is an item (file or directory) in repository root
    repository_root_items = []
    for elem in repository.iter():
        if elem.tag == "a" and not elem.text.startswith(".."):  # skip parent-dir if present
            repository_root_items.append(elem.attrib["href"])
    # Check if all treeinfo related directories are present
    # Trailing '/' is present for easier check
    for directory in RPM_KICKSTART_REPOSITORY_ROOT_CONTENT:
        assert directory in repository_root_items

    assert "repodata/" in repository_root_items
    # assert how many items are present altogether
    # here is '+1' for 'repodata'
    assert len(repository_root_items) == len(RPM_KICKSTART_REPOSITORY_ROOT_CONTENT) + 1


@pytest.fixture(scope="class")
def generate_distribution(
    init_and_sync,
    gen_object_with_cleanup,
    rpm_distribution_api,
    rpm_publication_api,
    monitor_task,
):
    def _generate_distribution(url=None):
        """Sync and publish an RPM repository.

        - create repository
        - create remote
        - sync the remote
        - create publication
        - create distribution

        Args:
            url(string):
                Optional URL of repository that should be use as a remote

        Returns (string):
            RPM distribution base_url.
        """
        repo, _ = init_and_sync(url=url)

        publish_data = RpmRpmPublication(repository=repo.pulp_href)
        publish_response = rpm_publication_api.create(publish_data)
        created_resources = monitor_task(publish_response.task).created_resources
        publication_href = created_resources[0]

        body = gen_distribution(publication=publication_href)
        distribution = gen_object_with_cleanup(rpm_distribution_api, body)

        return distribution.to_dict()["base_url"]

    return _generate_distribution



import pytest


@pytest.mark.parametrize(
    "factory_kwargs,repo_config_result",
    [
        ({}, {}),
        ({"gpgcheck": 0}, {"gpgcheck": 0}),
        ({"repo_gpgcheck": 0}, {"repo_gpgcheck": 0}),
        ({"gpgcheck": 0, "repo_gpgcheck": 0}, {"gpgcheck": 0, "repo_gpgcheck": 0}),
        ({"gpgcheck": 1, "repo_gpgcheck": 1}, {"gpgcheck": 1, "repo_gpgcheck": 1}),
    ],
)
def test_create_repo_with_deprecated_gpg_options_3357(
    rpm_repository_factory, factory_kwargs, repo_config_result
):
    """Can create repository with deprecated gpgcheck and repo_gpgcheck options."""
    repo = rpm_repository_factory(**factory_kwargs)
    assert repo.repo_config == repo_config_result


@pytest.mark.parametrize(
    "original_repo_config,update_kwargs,repo_config_result",
    [
        ({}, {}, {}),
        ({}, {"gpgcheck": 0}, {"gpgcheck": 0}),
        ({}, {"repo_gpgcheck": 0}, {"repo_gpgcheck": 0}),
        ({}, {"gpgcheck": 0, "repo_gpgcheck": 0}, {"gpgcheck": 0, "repo_gpgcheck": 0}),
        ({}, {"gpgcheck": 1, "repo_gpgcheck": 1}, {"gpgcheck": 1, "repo_gpgcheck": 1}),
        (
            {"gpgcheck": 0, "repo_gpgcheck": 0},
            {"gpgcheck": 1, "repo_gpgcheck": 1},
            {"gpgcheck": 1, "repo_gpgcheck": 1},
        ),
    ],
)
def test_update_repo_with_deprecated_gpg_options_3357(
    rpm_repository_factory,
    rpm_repository_api,
    original_repo_config,
    update_kwargs,
    repo_config_result,
):
    """Can update repository with deprecated gpgcheck and repo_gpgcheck options."""
    original_repo = rpm_repository_factory(description="old", **original_repo_config)
    assert original_repo.repo_config == original_repo_config
    assert original_repo.description == "old"  # control group

    body = {"name": original_repo.name}
    body.update(update_kwargs)
    body.update({"description": "new"})
    rpm_repository_api.update(original_repo.pulp_href, body)
    updated_repo = rpm_repository_api.read(original_repo.pulp_href)
    assert updated_repo.repo_config == repo_config_result
    assert updated_repo.description == "new"  # control group



from django.db import migrations

def delete_orphan_subrepos(apps, schema):
    """
    Remove subrepos which don't belong to any variant or addon.
    """
    RpmRepository = apps.get_model("rpm", "RpmRepository")
    RpmRepository.objects.filter(addons=None,variants=None,sub_repo=True).delete()

class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0024_change_subrepo_relation_properties'),
    ]

    operations = [
        migrations.RunPython(delete_orphan_subrepos)
    ]



# Generated by Django 2.2.12 on 2020-06-16 12:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0010_revision_null_redo'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmremote',
            name='sles_auth_token',
            field=models.CharField(max_length=512, null=True),
        ),
    ]



"""Tests that create/sync/distribute/publish MANY rpm plugin repositories."""

import os
import re

import pytest
import requests


@pytest.fixture
def cdn_certs_and_keys():
    cdn_client_cert = os.getenv("CDN_CLIENT_CERT", "").replace("\\n", "\n")
    cdn_client_key = os.getenv("CDN_CLIENT_KEY", "").replace("\\n", "\n")
    cdn_ca_cert = os.getenv("CDN_CA_CERT", "").replace("\\n", "\n")

    return cdn_client_cert, cdn_client_key, cdn_ca_cert


def _name_from_url(url):
    """Converts a remote-url into a string suitable for name-fields."""
    # https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/extras/os
    # cdn_redhat_com_content_dist_rhel_server_6_6Server_x86_64_extras_os

    # drop trailing slash
    rstr = url.rstrip("/")

    # drop protocol
    if rstr.startswith("https://"):
        rstr = rstr[8:]
    elif rstr.startswith("http://"):
        rstr = rstr[7:]

    # convert ./- into underscore
    rstr = re.sub("[.\-/]", "_", rstr)  # noqa
    return rstr


# 'export FIPS_WORKFLOW="anything"' to run this suite
@pytest.mark.skipif(
    not os.environ.get("FIPS_WORKFLOW", None),
    reason="This is a SIX HOUR test suit - run only when you're sure it's what you need",
)
@pytest.mark.parallel
@pytest.mark.parametrize(
    "url",
    [
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/extras/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/extras/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/rhscl/1/os",
        "http://mirror.centos.org/centos-7/7/extras/x86_64/",
        "http://mirror.centos.org/centos-7/7/sclo/x86_64/sclo/",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.6/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.7/x86_64/kickstart",
        "https://cdn.redhat.com/content/dist/rhel8/8.0/x86_64/baseos/kickstart",
        "https://mirrors.kernel.org/fedora-epel/7/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.7/x86_64/kickstart",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.6/x86_64/rhscl/1/os",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.6/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.3/x86_64/kickstart",
        "https://cdn.redhat.com/content/dist/rhel8/8.0/x86_64/appstream/kickstart",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.3/x86_64/optional/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.3/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.3/x86_64/rhscl/1/os",
        "http://vault.centos.org/6.10/os/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/ansible/2.5/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhgs-server-nfs/3.1/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.6/x86_64/kickstart",
        "https://cdn.redhat.com/content/dist/rhel/workstation/7/7.5/x86_64/kickstart",
        "https://mirrors.kernel.org/fedora-epel/8/Everything/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/workstation/7/7Workstation/x86_64/insights/3/os",
        "https://cdn.redhat.com/content/dist/rhel/workstation/7/7Workstation/x86_64/rh-common/os",
        "https://cdn.redhat.com/content/dist/rhel/workstation/7/7Workstation/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/workstation/7/7Workstation/x86_64/extras/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rh-gluster-samba/3.1/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhgs-server/3.1/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.10/x86_64/kickstart",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhgs-nagios/3.1/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhscon-agent/2/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhscon-main/2/os",
        "http://vault.centos.org/6.10/updates/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/rhs-client/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhs-client/os",
        "http://mirror.centos.org/centos-7/7/sclo/x86_64/rh/",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/dist/rhel/workstation/7/7.6/x86_64/kickstart",
        "http://mirror.centos.org/centos-7/7/updates/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.8/x86_64/kickstart",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.9/x86_64/kickstart",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.6/x86_64/sat-tools/6.2/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhscl/1/os",
        "http://mirror.centos.org/centos-7/7/os/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/ansible/2.7/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.3/x86_64/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.6/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.4/x86_64/kickstart",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.5/x86_64/kickstart",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhgs-server-bigdata/3.1/os",  # noqa E501
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/rhgs-server-splunk/3.1/os",  # noqa E501
        "https://cdn.redhat.com/content/dist/rhel/workstation/7/7Workstation/x86_64/supplementary/os",  # noqa E501
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/ansible/2.6/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/dotnet/1/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.10/x86_64/optional/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7Server/x86_64/sat-tools/6.5/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.5/x86_64/sat-tools/6.5/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.7/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.4/x86_64/optional/os",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.7/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.7/x86_64/optional/os",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.7/x86_64/rhscl/1/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.5/x86_64/rhscl/1/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.5/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.7/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.10/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.8/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.6/x86_64/os",
        "https://cdn.redhat.com/content/eus/rhel/server/6/6.7/x86_64/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.5/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.5/x86_64/sat-tools/6.4/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.5/x86_64/optional/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.3/x86_64/sat-tools/6.4/os",
        "https://cdn.redhat.com/content/dist/rhel8/8/x86_64/appstream/os",
        "https://cdn.redhat.com/content/dist/rhel8/8/x86_64/baseos/os",
        "https://cdn.redhat.com/content/dist/rhel8/8/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/dist/rhel8/8/x86_64/baseos/kickstart",
        "https://cdn.redhat.com/content/dist/rhel8/8/x86_64/appstream/kickstart",
        "https://archives.fedoraproject.org/pub/archive/epel/6/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.6/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.3/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.9/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.8/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.7/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6.9/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.5/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.2/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.6/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.5/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.3/x86_64/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.6/x86_64/sat-tools/6.5/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.6/x86_64/optional/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/sat-capsule/6.6/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/sat-tools/6.6/os",
        "https://cdn.redhat.com/content/dist/layered/rhel8/x86_64/sat-tools/6.6/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/ansible/2.8/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.6/x86_64/sat-tools/6.6/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.7/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel8/8.1/x86_64/appstream/kickstart",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.4/x86_64/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.2/x86_64/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.6/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.6/x86_64/rhscl/1/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/sat-tools/6.6/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.5/x86_64/sat-tools/6.6/os",
        "https://cdn.redhat.com/content/dist/rhel8/8.1/x86_64/baseos/kickstart",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/highavailability/os",
        "https://packages.vmware.com/tools/releases/10.3.5/rhel6/x86_64/",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7.8/x86_64/kickstart",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.7/x86_64/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.7/x86_64/supplementary/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.7/x86_64/rhscl/1/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.7/x86_64/optional/os",
        "https://cdn.redhat.com/content/eus/rhel/server/7/7.7/x86_64/sat-tools/6.6/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/sat-capsule/6.7/os",
        "https://cdn.redhat.com/content/dist/rhel/server/6/6Server/x86_64/sat-tools/6.7/os",
        "https://cdn.redhat.com/content/dist/rhel/server/7/7Server/x86_64/sat-tools/6.7/os",
        "https://cdn.redhat.com/content/dist/layered/rhel8/x86_64/sat-tools/6.7/os",
        "https://cdn.redhat.com/content/dist/rhel8/8.2/x86_64/appstream/kickstart",
        "https://cdn.redhat.com/content/dist/rhel8/8.2/x86_64/baseos/kickstart",
        "http://mirror.centos.org/centos-8/8/BaseOS/x86_64/os/",
        "http://mirror.centos.org/centos-8/8/AppStream/x86_64/os/",
    ],
)
def test_fips_workflow(
    url,
    init_and_sync,
    rpm_rpmremote_factory,
    rpm_publication_factory,
    rpm_distribution_factory,
    cdn_certs_and_keys,
):
    # Convert a url into a name-string
    name = _name_from_url(url)

    # Create a remote
    body = {"url": url, "policy": "on_demand"}
    if name.startswith("cdn_"):
        if not all(cdn_certs_and_keys):
            pytest.skip("Can't test CDN repositories, missing certs")
        # need cert-access
        client_cert, client_key, ca_cert = cdn_certs_and_keys
        body["ca_cert"] = ca_cert
        body["client_cert"] = client_cert
        body["client_key"] = client_key
    remote = rpm_rpmremote_factory(**body)
    assert remote is not None

    # Create a repo & sync w/ remote
    repo, _ = init_and_sync(remote=remote)

    # Publish the result
    publication = rpm_publication_factory(repository=repo.pulp_href)
    assert publication is not None

    # Distribute the published version
    distribution = rpm_distribution_factory(publication=publication.pulp_href)
    assert distribution is not None

    # Test we can access the index of the distribution
    response = requests.get(distribution.base_url)
    assert response is not None



from pulpcore.plugin.viewsets import (
    ReadOnlyContentViewSet,
)

from pulp_rpm.app.models import (
    DistributionTree,
)
from pulp_rpm.app.serializers import (
    DistributionTreeSerializer,
)


class DistributionTreeViewSet(ReadOnlyContentViewSet):
    """
    Distribution Tree Viewset.

    """

    endpoint_name = "distribution_trees"
    queryset = DistributionTree.objects.all()
    serializer_class = DistributionTreeSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }



import pytest

from pulpcore.tests.functional.utils import PulpTaskError

from pulp_rpm.tests.functional.constants import (
    PULP_FIXTURES_BASE_URL,
    RPM_FIXTURE_SUMMARY,
    RPM_KICKSTART_FIXTURE_SUMMARY,
    RPM_KICKSTART_ONLY_META_FIXTURE_URL,
    RPM_ONLY_METADATA_REPO_URL,
)

from pulpcore.client.pulp_rpm import (
    RpmRepositorySyncURL,
)


@pytest.mark.parametrize(
    "paths,remote_url,content_summary",
    [
        (["rpm-unsigned/"], RPM_ONLY_METADATA_REPO_URL, RPM_FIXTURE_SUMMARY),
        (
            ["rpm-distribution-tree/"],
            RPM_KICKSTART_ONLY_META_FIXTURE_URL,
            RPM_KICKSTART_FIXTURE_SUMMARY,
        ),
    ],
)
def test_acs_simple(
    paths,
    remote_url,
    content_summary,
    rpm_repository_version_api,
    rpm_repository_api,
    rpm_acs_api,
    rpm_repository_factory,
    rpm_rpmremote_factory,
    monitor_task,
    monitor_task_group,
    gen_object_with_cleanup,
    delete_orphans_pre,
):
    """Test to sync repo with use of ACS."""
    # ACS is rpm-unsigned repository which has all packages needed
    acs_remote = rpm_rpmremote_factory(url=PULP_FIXTURES_BASE_URL, policy="on_demand")

    acs_data = {
        "name": "alternatecontentsource",
        "remote": acs_remote.pulp_href,
        "paths": paths,
    }
    acs = gen_object_with_cleanup(rpm_acs_api, acs_data)

    repo = rpm_repository_factory()
    remote = rpm_rpmremote_factory(url=remote_url)

    # Sync repo with metadata only, before ACS refresh it should fail
    repository_sync_data = RpmRepositorySyncURL(remote=remote.pulp_href)

    with pytest.raises(PulpTaskError) as ctx:
        sync_response = rpm_repository_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)

    assert "404, message='Not Found'" in ctx.value.task.error["description"]

    # ACS refresh
    acs_refresh = rpm_acs_api.refresh(acs.pulp_href)
    monitor_task_group(acs_refresh.task_group)

    # Sync repository with metadata only
    sync_response = rpm_repository_api.sync(repo.pulp_href, repository_sync_data)
    monitor_task(sync_response.task)

    repo = rpm_repository_api.read(repo.pulp_href)
    repo_ver = rpm_repository_version_api.read(repo.latest_version_href)

    present_summary = {k: v["count"] for k, v in repo_ver.content_summary.present.items()}
    assert present_summary == content_summary



import logging

import createrepo_c as cr

log = logging.getLogger(__name__)


def warningcb(warning_type, message):
    """Optional callback for warnings about wierd stuff and formatting in XML.

    Args:
        warning_type (int): One of the XML_WARNING_* constants.
        message (str): Message.
    """
    log.warn("PARSER WARNING: %s" % message)
    return True  # continue parsing


class MetadataParser:
    """Parser for RPM metadata."""

    def __init__(self):
        """Initialize empty (use one of the alternate constructors)."""
        self.primary_xml_path = None
        self.filelists_xml_path = None
        self.other_xml_path = None

    @staticmethod
    def from_metadata_files(primary_xml_path, filelists_xml_path, other_xml_path):
        """Construct a parser from the three main metadata files."""
        parser = MetadataParser()
        parser.primary_xml_path = primary_xml_path
        parser.filelists_xml_path = filelists_xml_path
        parser.other_xml_path = other_xml_path
        return parser

    def for_each_pkg_primary(self, pkgcb):
        """Execute a callback for each package, parsed from only primary package metadata.

        Only primary metadata means no files or changelogs.
        """
        cr.xml_parse_primary(self.primary_xml_path, pkgcb=pkgcb, do_files=False)

    def as_iterator(self):
        """Return a package iterator."""
        return cr.PackageIterator(
            primary_path=self.primary_xml_path,
            filelists_path=self.filelists_xml_path,
            other_path=self.other_xml_path,
            warningcb=warningcb,
        )



from gettext import gettext as _
from rest_framework import serializers

from pulpcore.plugin.serializers import (
    ContentChecksumSerializer,
    SingleArtifactContentUploadSerializer,
)
from pulpcore.plugin.util import get_domain_pk

from pulp_rpm.app.models import RepoMetadataFile


class RepoMetadataFileSerializer(SingleArtifactContentUploadSerializer, ContentChecksumSerializer):
    """
    RepoMetadataFile serializer.
    """

    data_type = serializers.CharField(help_text=_("Metadata type."))
    checksum_type = serializers.CharField(help_text=_("Checksum type for the file."))
    checksum = serializers.CharField(help_text=_("Checksum for the file."))
    relative_path = serializers.CharField(help_text=_("Relative path of the file."))

    # Mirror unique_together from the Model
    def retrieve(self, validated_data):
        content = RepoMetadataFile.objects.filter(
            data_type=validated_data["data_type"],
            checksum=validated_data["checksum"],
            relative_path=validated_data["relative_path"],
            pulp_domain=get_domain_pk(),
        )
        return content.first()

    class Meta:
        fields = (
            ContentChecksumSerializer.Meta.fields
            + SingleArtifactContentUploadSerializer.Meta.fields
            + (
                "data_type",
                "checksum_type",
                "checksum",
            )
        )
        model = RepoMetadataFile



# Generated by Django 3.2.9 on 2021-12-20 22:05

from django.db import migrations, models


def generate_fake_digest(apps, schema_editor):
    """
    Generate a temporary digest for a dist tree

    This is needed so the dist tree is fixed by a subsequent sync.
    We can't use digest from .treeinfo file because a dist tree object is potentially wrong and
    we need to let a subsequent [force] sync to fix it.
    """
    DistributionTree = apps.get_model('rpm', 'DistributionTree')
    dist_trees = DistributionTree.objects.all()
    for idx, dist_tree in enumerate(dist_trees):
        dist_tree.digest = 'temp str {0:032d}, to be removed at sync'.format(idx)
    DistributionTree.objects.bulk_update(dist_trees, fields=['digest'], batch_size=500)


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0038_fix_sync_optimization'),
    ]

    operations = [
        migrations.AddField(
            model_name='distributiontree',
            name='digest',
            field=models.CharField(default='', max_length=64),
            preserve_default=False,
        ),
        migrations.RunPython(
            generate_fake_digest
        ),
        migrations.AlterUniqueTogether(
            name='distributiontree',
            unique_together={('digest',)},
        ),
    ]



# Generated by Django 4.2.4 on 2023-09-08 14:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("rpm", "0054_remove_gpg_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="rpmpublication",
            name="repo_config",
            field=models.JSONField(default=dict),
        ),
        migrations.AddField(
            model_name="rpmrepository",
            name="repo_config",
            field=models.JSONField(default=dict),
        ),
    ]



from django.db import migrations, transaction
from django.core.exceptions import ObjectDoesNotExist

def remove_publications_from_auto_distributed(apps, schema_editor):
    with transaction.atomic():
        RpmDistribution = apps.get_model("rpm", "RpmDistribution")
        distributions = RpmDistribution.objects.filter(repository__isnull=False, publication__isnull=False)
        distributions.update(publication=None)

def add_publications_to_auto_distributed(apps, schema_editor):
    with transaction.atomic():
        RpmDistribution = apps.get_model("rpm", "RpmDistribution")
        distributions = list(RpmDistribution.objects.filter(repository__isnull=False).select_related("repository"))
        for distribution in distributions:
            repo_version = distribution.repository.latest_version()
            try:
                publication = repo_version.publication_set.earliest("pulp_created")
            except ObjectDoesNotExist:
                publication = None
            distribution.publication = publication
        RpmDistribution.objects.bulk_update(distributions, ['publication'])


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0034_auto_publish'),
    ]

    operations = [
        migrations.RunPython(
            remove_publications_from_auto_distributed,
            reverse_code=add_publications_to_auto_distributed
        )
    ]



# Generated by Django 4.2.4 on 2023-09-04 14:06

from django.db import migrations, models
import hashlib


def add_snippet_hash(apps, schema_editor):
    """Calculate and add digest hash of the snippet."""

    Modulemd = apps.get_model("rpm", "Modulemd")
    modules_to_update = []

    for mmd in Modulemd.objects.all().only("snippet").iterator():
        # for the modules that have empty string snippet will be calculated digest of
        # e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
        mmd.digest = hashlib.sha256(mmd.snippet.encode()).hexdigest()
        modules_to_update.append(mmd)
    Modulemd.objects.bulk_update(modules_to_update, fields=["digest"])


class Migration(migrations.Migration):

    dependencies = [
        ("rpm", "0051_alter_distributiontree_unique_together_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="modulemd",
            name="digest",
            field=models.TextField(null=True),
        ),
        migrations.RunPython(
            code=add_snippet_hash,
            reverse_code=migrations.RunPython.noop,
            elidable=True,
        ),
        migrations.AlterField(
            model_name="modulemd",
            name="digest",
            field=models.TextField(),
        ),
        migrations.AlterUniqueTogether(
            name="modulemd",
            unique_together=set(),
        ),
        migrations.AlterUniqueTogether(
            name="modulemd",
            unique_together={
                ("_pulp_domain", "name", "stream", "version", "context", "arch", "digest")
            },
        ),
    ]



import pytest
import requests


@pytest.fixture
def setup_empty_distribution(
    rpm_repository_factory, rpm_publication_factory, rpm_distribution_factory
):
    repo = rpm_repository_factory()
    publication = rpm_publication_factory(repository=repo.pulp_href)
    distribution = rpm_distribution_factory(
        publication=publication.pulp_href, generate_repo_config=True
    )

    return repo, publication, distribution


@pytest.mark.parallel
def test_config_repo_in_listing_unsigned(setup_empty_distribution):
    """Whether the served resources are in the directory listing."""
    _, _, dist = setup_empty_distribution
    content = requests.get(dist.base_url).content

    assert b"config.repo" in content
    assert b"repomd.xml.key" not in content


@pytest.mark.parallel
def test_config_repo_unsigned(setup_empty_distribution):
    """Whether config.repo can be downloaded and has the right content."""
    _, _, dist = setup_empty_distribution
    content = requests.get(f"{dist.base_url}config.repo").content

    assert bytes(f"[{dist.name}]\n", "utf-8") in content
    assert bytes(f"baseurl={dist.base_url}\n", "utf-8") in content
    assert bytes("gpgcheck=0\n", "utf-8") in content
    assert bytes("repo_gpgcheck=0", "utf-8") in content


@pytest.mark.parallel
def test_config_repo_auto_distribute(
    setup_empty_distribution, rpm_publication_api, rpm_distribution_api, monitor_task
):
    """Whether config.repo is properly served using auto-distribute."""
    repo, pub, dist = setup_empty_distribution

    body = {"repository": repo.pulp_href, "publication": None}
    monitor_task(rpm_distribution_api.partial_update(dist.pulp_href, body).task)
    # Check that distribution is now using repository to auto-distribute
    dist = rpm_distribution_api.read(dist.pulp_href)
    assert repo.pulp_href == dist.repository
    assert dist.publication is None
    content = requests.get(f"{dist.base_url}config.repo").content

    assert bytes(f"[{dist.name}]\n", "utf-8") in content
    assert bytes(f"baseurl={dist.base_url}\n", "utf-8") in content
    assert bytes("gpgcheck=0\n", "utf-8") in content
    assert bytes("repo_gpgcheck=0", "utf-8") in content

    # Delete publication and check that 404 is now returned
    rpm_publication_api.delete(pub.pulp_href)
    assert requests.get(f"{dist.base_url}config.repo").status_code == 404



"""Tests for the retention policy feature of repositories."""

import pytest

from collections import defaultdict

from pulp_rpm.tests.functional.constants import (
    PULP_TYPE_PACKAGE,
    RPM_FIXTURE_SUMMARY,
    RPM_PACKAGE_COUNT,
    RPM_MODULAR_PACKAGE_COUNT,
    RPM_MODULES_STATIC_CONTEXT_FIXTURE_URL,
    RPM_MODULAR_STATIC_FIXTURE_SUMMARY,
)
from pulpcore.client.pulp_rpm.exceptions import ApiException


def test_sync_with_retention(
    delete_orphans_pre,
    init_and_sync,
    rpm_repository_api,
    rpm_repository_version_api,
    rpm_package_api,
    monitor_task,
):
    """Verify functionality with sync.

    Do the following:

    1. Create a repository, and a remote.
    2. Sync the remote.
    3. Assert that the correct number of units were added and are present
       in the repo.
    4. Change the "retain_package_versions" on the repository to 1 (retain the latest
       version only).
    5. Sync the remote one more time.
    6. Assert that repository version is different from the previous one.
    7. Assert the repository version we end with has only one version of each package.
    """
    repo, remote, task = init_and_sync(policy="on_demand", optimize=False, return_task=True)

    # Test that, by default, everything is retained / nothing is tossed out.
    version = rpm_repository_version_api.read(repo.latest_version_href)
    present = {k: v["count"] for k, v in version.content_summary.present.items()}
    added = {k: v["count"] for k, v in version.content_summary.added.items()}
    assert present == RPM_FIXTURE_SUMMARY
    assert added == RPM_FIXTURE_SUMMARY
    # Test that the # of packages processed is correct
    reports = get_progress_reports_by_code(task)
    assert reports["sync.parsing.packages"].total == RPM_PACKAGE_COUNT

    # Set the retention policy to retain only 1 version of each package
    repo_data = repo.to_dict()
    repo_data.update({"retain_package_versions": 1})
    monitor_task(rpm_repository_api.update(repo.pulp_href, repo_data).task)
    repo = rpm_repository_api.read(repo.pulp_href)

    repo, remote, task = init_and_sync(
        repository=repo, remote=remote, optimize=False, return_task=True
    )

    # Test that only one version of each package is present
    content = rpm_package_api.list(repository_version=repo.latest_version_href).results
    assert check_retention_policy(content, 1)
    # Test that (only) 4 RPMs were removed (no advisories etc. touched)
    version = rpm_repository_version_api.read(repo.latest_version_href)
    removed = {k: v["count"] for k, v in version.content_summary.removed.items()}
    assert removed == {PULP_TYPE_PACKAGE: 4}
    # Test that the versions that were removed are the versions we expect.
    content = rpm_package_api.list(repository_version_removed=repo.latest_version_href).results
    versions = versions_for_packages(content)
    assert versions == {"duck": ["0.6", "0.7"], "kangaroo": ["0.2"], "walrus": ["0.71"]}, versions
    # Test that the number of packages processed is correct (doesn't include older ones)
    reports = get_progress_reports_by_code(task)
    assert reports["sync.parsing.packages"].total == RPM_PACKAGE_COUNT
    assert reports["sync.skipped.packages"].total == 4


def test_sync_with_retention_and_modules(
    delete_orphans_pre,
    init_and_sync,
    rpm_repository_api,
    rpm_repository_version_api,
    monitor_task,
):
    """Verify functionality with sync.

    Do the following:

    1. Create a repository, and a remote.
    2. Sync the remote.
    3. Assert that the correct number of units were added and are present in the repo.
    4. Change the "retain_package_versions" on the repository to 1 (retain the latest
       version only).
    5. Sync the remote one more time.
    6. Assert that repository version is the same as the previous one, because the older
       versions are part of modules, and they should be ignored by the retention policy.
    """

    repo, remote, task = init_and_sync(
        url=RPM_MODULES_STATIC_CONTEXT_FIXTURE_URL,
        policy="on_demand",
        optimize=False,
        return_task=True,
    )

    # Test that, by default, everything is retained / nothing is tossed out.
    version = rpm_repository_version_api.read(repo.latest_version_href)
    present = {k: v["count"] for k, v in version.content_summary.present.items()}
    added = {k: v["count"] for k, v in version.content_summary.added.items()}
    assert present == RPM_MODULAR_STATIC_FIXTURE_SUMMARY
    assert added == RPM_MODULAR_STATIC_FIXTURE_SUMMARY
    # Test that the # of packages processed is correct
    reports = get_progress_reports_by_code(task)
    assert reports["sync.parsing.packages"].total == RPM_MODULAR_PACKAGE_COUNT
    assert reports["sync.skipped.packages"].total == 0

    # Set the retention policy to retain only 1 version of each package
    repo_data = repo.to_dict()
    repo_data.update({"retain_package_versions": 1})
    monitor_task(rpm_repository_api.update(repo.pulp_href, repo_data).task)
    repo = rpm_repository_api.read(repo.pulp_href)

    repo, remote, task = init_and_sync(
        repository=repo, remote=remote, optimize=False, return_task=True
    )

    # Test that no RPMs were removed (and no advisories etc. touched)
    # it should be the same because the older version are covered by modules
    version = rpm_repository_version_api.read(repo.latest_version_href)
    assert version.content_summary.removed == {}
    # Test that the number of packages processed is correct
    reports = get_progress_reports_by_code(task)
    assert reports["sync.parsing.packages"].total == RPM_MODULAR_PACKAGE_COUNT
    assert reports["sync.skipped.packages"].total == 0


@pytest.mark.parallel
def test_mirror_sync_with_retention_fails(init_and_sync, rpm_repository_factory):
    """Verify functionality with sync.

    Do the following:

    1. Create a repository with 'retain_package_versions' set, and a remote.
    2. Sync the remote in mirror mode.
    3. Assert that the sync fails.
    """
    repo = rpm_repository_factory(retain_package_versions=1)
    with pytest.raises(ApiException) as exc:
        init_and_sync(repository=repo, optimize=False, sync_policy="mirror_complete")

    assert exc.value.status == 400


def get_progress_reports_by_code(task):
    """Return the progress reports in a dictionary keyed by codename."""
    return {report.code: report for report in task.progress_reports}


def versions_for_packages(packages):
    """Get a list of versions for each package present in a list of Package dicts.

    Args:
        packages: List of Package info dicts
    """
    packages_by_version = defaultdict(list)

    for package in packages:
        packages_by_version[package.name].append(package.version)

    for pkg_list in packages_by_version.values():
        pkg_list.sort()

    return packages_by_version


def check_retention_policy(packages, retain_package_versions):
    """Check that the number of versions of each package <= permitted number.

    Args:
        packages: List of Package info dicts
        retention_policy: Number of package versions permitted.
    """
    return all(
        [
            len(versions) <= retain_package_versions
            for versions in versions_for_packages(packages).values()
        ]
    )



import re
import os
import textwrap

from gettext import gettext as _
from logging import getLogger

from aiohttp.web_response import Response
from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings
from django.db import models
from pulpcore.plugin.download import DownloaderFactory
from pulpcore.plugin.models import (
    AutoAddObjPermsMixin,
    Artifact,
    AsciiArmoredDetachedSigningService,
    Content,
    ContentArtifact,
    Remote,
    RemoteArtifact,
    Repository,
    RepositoryContent,
    RepositoryVersion,
    Publication,
    PublishedMetadata,
    Distribution,
)
from pulpcore.plugin.repo_version_utils import (
    remove_duplicates,
    validate_duplicate_content,
    validate_version_paths,
)

from pulp_rpm.app.constants import CHECKSUM_CHOICES, COMPRESSION_CHOICES
from pulp_rpm.app.models import (
    DistributionTree,
    Package,
    PackageCategory,
    PackageGroup,
    PackageEnvironment,
    PackageLangpacks,
    RepoMetadataFile,
    Modulemd,
    ModulemdDefaults,
    ModulemdObsolete,
    UpdateRecord,
)

from pulp_rpm.app.downloaders import RpmDownloader, RpmFileDownloader, UlnDownloader
from pulp_rpm.app.exceptions import DistributionTreeConflict
from pulp_rpm.app.shared_utils import urlpath_sanitize

log = getLogger(__name__)


class RpmRemote(Remote, AutoAddObjPermsMixin):
    """
    Remote for "rpm" content.
    """

    TYPE = "rpm"
    sles_auth_token = models.TextField(null=True)

    DEFAULT_DOWNLOAD_CONCURRENCY = 7
    DEFAULT_MAX_RETRIES = 4

    @property
    def download_factory(self):
        """
        Return the DownloaderFactory which can be used to generate asyncio capable downloaders.

        Returns:
            DownloadFactory: The instantiated DownloaderFactory to be used by
                get_downloader()

        """
        try:
            return self._download_factory
        except AttributeError:
            self._download_factory = DownloaderFactory(
                self,
                downloader_overrides={
                    "http": RpmDownloader,
                    "https": RpmDownloader,
                    "file": RpmFileDownloader,
                },
            )
            return self._download_factory

    def get_downloader(self, remote_artifact=None, url=None, **kwargs):
        """
        Get a downloader from either a RemoteArtifact or URL that is configured with this Remote.

        This method accepts either `remote_artifact` or `url` but not both. At least one is
        required. If neither or both are passed a ValueError is raised.

        Args:
            remote_artifact (:class:`~pulpcore.app.models.RemoteArtifact`): The RemoteArtifact to
                download.
            url (str): The URL to download.
            kwargs (dict): This accepts the parameters of
                :class:`~pulpcore.plugin.download.BaseDownloader`.
        Raises:
            ValueError: If neither remote_artifact and url are passed, or if both are passed.
        Returns:
            subclass of :class:`~pulpcore.plugin.download.BaseDownloader`: A downloader that
            is configured with the remote settings.

        """
        if self.sles_auth_token:
            kwargs["sles_auth_token"] = self.sles_auth_token
        return super().get_downloader(remote_artifact=remote_artifact, url=url, **kwargs)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [
            ("manage_roles_rpmremote", "Can manage roles on an RPM remotes"),
        ]


class UlnRemote(Remote, AutoAddObjPermsMixin):
    """
    Remote for "uln" content.
    """

    TYPE = "uln"
    uln_server_base_url = models.TextField(null=True)

    @property
    def download_factory(self):
        """
        Return the DownloaderFactory which can be used to generate asyncio capable downloaders.

        Returns:
            DownloadFactory: The instantiated DownloaderFactory to be used by
                get_downloader()

        """
        try:
            return self._download_factory
        except AttributeError:
            self._download_factory = DownloaderFactory(
                self,
                downloader_overrides={
                    "uln": UlnDownloader,
                },
            )
            self._download_factory._handler_map["uln"] = self._download_factory._http_or_https
            return self._download_factory

    def get_downloader(self, remote_artifact=None, url=None, **kwargs):
        """
        Get a downloader from either a RemoteArtifact or URL that is configured with this Remote.

        This method accepts either `remote_artifact` or `url` but not both. At least one is
        required. If neither or both are passed a ValueError is raised.

        Args:
            remote_artifact (:class:`~pulpcore.app.models.RemoteArtifact`): The RemoteArtifact to
                download.
            url (str): The URL to download. Can be a ULN url.
            kwargs (dict): This accepts the parameters of
                :class:`~pulpcore.plugin.download.BaseDownloader`.
        Raises:
            ValueError: If neither remote_artifact and url are passed, or if both are passed.
        Returns:
            subclass of :class:`~pulpcore.plugin.download.BaseDownloader`: A downloader that
            is configured with the remote settings.

        """
        if self.uln_server_base_url:
            uln_server_base_url = self.uln_server_base_url
        else:
            uln_server_base_url = settings.DEFAULT_ULN_SERVER_BASE_URL

        return super().get_downloader(
            remote_artifact=remote_artifact,
            url=url,
            username=self.username,
            password=self.password,
            uln_server_base_url=uln_server_base_url,
            **kwargs,
        )

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [
            ("manage_roles_ulnremote", "Can manage roles on an ULN remotes"),
        ]


class RpmRepository(Repository, AutoAddObjPermsMixin):
    """
    Repository for "rpm" content.

    Fields:
        original_checksum_types (JSON): Checksum for each metadata type
        last_sync_details (JSON): Details about the last sync including repomd, settings used, etc.
        retain_package_versions (Integer): Max number of latest versions of each package to keep.
        autopublish (Boolean): Whether to automatically create a publication for new versions.
        metadata_checksum_type (String):
            The name of a checksum type to use for metadata when generating metadata.
        package_checksum_type (String):
            The name of a default checksum type to use for packages when generating metadata.
        repo_config (JSON): repo configuration that will be served by distribution
        compression_type(pulp_rpm.app.constants.COMPRESSION_TYPES):
            Compression type to use for metadata files.
    """

    TYPE = "rpm"
    CONTENT_TYPES = [
        Package,
        UpdateRecord,
        PackageCategory,
        PackageGroup,
        PackageEnvironment,
        PackageLangpacks,
        RepoMetadataFile,
        DistributionTree,
        Modulemd,
        ModulemdDefaults,
        ModulemdObsolete,
    ]
    REMOTE_TYPES = [RpmRemote, UlnRemote]

    metadata_signing_service = models.ForeignKey(
        AsciiArmoredDetachedSigningService, on_delete=models.SET_NULL, null=True
    )
    original_checksum_types = models.JSONField(default=dict)
    last_sync_details = models.JSONField(default=dict)
    retain_package_versions = models.PositiveIntegerField(default=0)

    autopublish = models.BooleanField(default=False)
    checksum_type = models.TextField(null=True, choices=CHECKSUM_CHOICES)
    compression_type = models.TextField(null=True, choices=COMPRESSION_CHOICES)
    metadata_checksum_type = models.TextField(null=True, choices=CHECKSUM_CHOICES)
    package_checksum_type = models.TextField(null=True, choices=CHECKSUM_CHOICES)
    repo_config = models.JSONField(default=dict)

    def on_new_version(self, version):
        """
        Called when new repository versions are created.

        Args:
            version: The new repository version.
        """
        super().on_new_version(version)

        # avoid circular import issues
        from pulp_rpm.app import tasks

        if self.autopublish:
            tasks.publish(
                repository_version_pk=version.pk,
                metadata_signing_service=self.metadata_signing_service,
                checksum_types={
                    "general": self.checksum_type,
                    "metadata": self.metadata_checksum_type,
                    "package": self.package_checksum_type,
                },
                repo_config=self.repo_config,
                compression_type=self.compression_type,
            )

    @property
    def published_metadata_size(self):
        versions = self.versions.all()
        publications = Publication.objects.filter(repository_version__in=versions, complete=True)
        published_metadata = PublishedMetadata.objects.filter(publication__in=publications)
        size = (
            Artifact.objects.filter(content__in=published_metadata)
            .distinct()
            .aggregate(size=models.Sum("size", default=0))["size"]
        )
        return size

    def all_content_pks(self):
        """Returns a list of pks for all content stored across all versions."""
        all_content = (
            RepositoryContent.objects.filter(repository=self)
            .distinct("content")
            .values_list("content")
        )
        repos = {self.pk}
        for dt in DistributionTree.objects.only().filter(pk__in=all_content):
            repos.update(dt.repositories().values_list("pk", flat=True))
        return (
            RepositoryContent.objects.filter(repository__in=repos)
            .distinct("content")
            .values_list("content")
        )

    @property
    def disk_size(self):
        """Returns the approximate size on disk for all artifacts stored across all versions."""
        return (
            Artifact.objects.filter(content__in=self.all_content_pks())
            .distinct()
            .aggregate(size=models.Sum("size", default=0))["size"]
        )

    @property
    def on_demand_size(self):
        """Returns the approximate size of all on-demand artifacts stored across all versions."""
        on_demand_ca = ContentArtifact.objects.filter(
            content__in=self.all_content_pks(), artifact=None
        )
        # Aggregate does not work with distinct("fields") so sum must be done manually
        ras = RemoteArtifact.objects.filter(
            content_artifact__in=on_demand_ca, size__isnull=False
        ).distinct("content_artifact")
        return sum(ras.values_list("size", flat=True))

    @staticmethod
    def on_demand_artifacts_for_version(version):
        """
        Returns the remote artifacts of on-demand content for a repository version.

        Override the default behavior to include DistributionTree artifacts from nested repos.
        Note: this only returns remote artifacts that have a non-null size.

        Args:
            version (pulpcore.app.models.RepositoryVersion): to get the remote artifacts for.
        Returns:
            django.db.models.QuerySet: The remote artifacts that are contained within this version.
        """
        content_pks = set(version.content.values_list("pk", flat=True))
        for tree in DistributionTree.objects.filter(pk__in=content_pks):
            content_pks.update(tree.content().values_list("pk", flat=True))
        on_demand_ca = ContentArtifact.objects.filter(content__in=content_pks, artifact=None)
        return RemoteArtifact.objects.filter(content_artifact__in=on_demand_ca, size__isnull=False)

    @staticmethod
    def artifacts_for_version(version):
        """
        Return the artifacts for an RpmRepository version.

        Override the default behavior to include DistributionTree artifacts from nested repos.

        Args:
            version (pulpcore.app.models.RepositoryVersion): to get the artifacts for

        Returns:
            django.db.models.QuerySet: The artifacts that are contained within this version.

        """
        artifacts_pk = set(
            Artifact.objects.filter(content__pk__in=version.content).values_list(
                "pulp_id", flat=True
            )
        )
        for tree in DistributionTree.objects.filter(pk__in=version.content):
            artifacts_pk |= set(tree.artifacts().values_list("pulp_id", flat=True))

        return Artifact.objects.filter(pk__in=artifacts_pk)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [
            ("manage_roles_rpmrepository", "Can manage roles on RPM repositories"),
            ("modify_content_rpmrepository", "Add content to, or remove content from a repository"),
            ("repair_rpmrepository", "Copy a repository"),
            ("sync_rpmrepository", "Sync a repository"),
            ("delete_rpmrepository_version", "Delete a repository version"),
        ]

    def finalize_new_version(self, new_version):
        """
        Ensure there are no duplicates in a repo version and content is not broken.

        Remove duplicates based on repo_key_fields.
        Ensure that modulemd is added with all its RPMs.
        Ensure that modulemd is removed with all its RPMs.
        Resolve advisory conflicts when there is more than one advisory with the same id.

        Args:
            new_version (pulpcore.app.models.RepositoryVersion): The incomplete RepositoryVersion
                to finalize.
        """
        if new_version.base_version:
            previous_version = new_version.base_version
        else:
            try:
                previous_version = new_version.previous()
            except RepositoryVersion.DoesNotExist:
                previous_version = None

        remove_duplicates(new_version)
        self._resolve_distribution_trees(new_version, previous_version)

        from pulp_rpm.app.modulemd import resolve_module_packages  # avoid circular import

        resolve_module_packages(new_version, previous_version)

        self._apply_retention_policy(new_version)

        from pulp_rpm.app.advisory import resolve_advisories  # avoid circular import

        resolve_advisories(new_version, previous_version)

        #
        # Some repositories are odd. A given NEVRA with different checksums can appear at
        # different locations in the repo, or a single Artifact can be referenced by more than one
        # name.
        #
        # validate_duplicate_content() takes repo-keys into account - so same-NEVRA, diff-location
        # passes the test.
        #
        # The validate_version_paths() test checks for different-nevras, but same relative-path,
        # and raises an exception. Because of these odd repositories, this can't be fatal - so
        # we warn about it, but continue. At publish, we will have to pick one.
        validate_duplicate_content(new_version)
        try:
            validate_version_paths(new_version)
        except ValueError as ve:
            log.warning(
                _(
                    "New version of repository {repo} reports duplicate/overlap errors : "
                    "{value_errors}"
                ).format(repo=new_version.repository.name, value_errors=str(ve))
            )

    def _apply_retention_policy(self, new_version):
        """Apply the repository's "retain_package_versions" settings to the new version.

        Remove all non-modular packages that are older than the retention policy. A value of 0
        for the package retention policy represents disabled. A value of 3 would mean that the
        3 most recent versions of each package would be kept while older versions are discarded.

        Args:
            new_version (models.RepositoryVersion): Repository version to filter
        """
        assert (
            not new_version.complete
        ), "Cannot apply retention policy to completed repository versions"

        if self.retain_package_versions > 0:
            # It would be more ideal if, instead of annotating with an age and filtering manually,
            # we could use Django to filter the particular Package content we want to delete.
            # Something like ".filter(F('age') > self.retain_package_versions)" would be better
            # however this is not currently possible with Django. It would be possible with raw
            # SQL but the repository version content membership subquery is currently
            # django-managed and would be difficult to share.
            #
            # Instead we have to do the filtering manually.
            nonmodular_packages = (
                Package.objects.with_age()
                .filter(
                    pk__in=new_version.content.filter(pulp_type=Package.get_pulp_type()),
                    is_modular=False,  # don't want to filter out modular RPMs
                )
                .only("pk")
            )

            old_packages = []
            for package in nonmodular_packages:
                if package.age > self.retain_package_versions:
                    old_packages.append(package.pk)

            new_version.remove_content(Content.objects.filter(pk__in=old_packages))

    def _resolve_distribution_trees(self, new_version, previous_version):
        """
        There can be only one distribution tree in a repo version.

        Args:
            version (pulpcore.app.models.RepositoryVersion): current incomplete repository version
            previous_version (pulpcore.app.models.RepositoryVersion):  a version preceding
                                                                       the current incomplete one
        """
        disttree_pulp_type = DistributionTree.get_pulp_type()
        current_disttrees = new_version.content.filter(pulp_type=disttree_pulp_type)

        if len(current_disttrees) < 2:
            return

        if previous_version:
            previous_disttree = previous_version.content.get(pulp_type=disttree_pulp_type)
            new_version.remove_content(Content.objects.filter(pk=previous_disttree.pk))

        incoming_disttrees = new_version.content.filter(pulp_type=disttree_pulp_type)
        if len(incoming_disttrees) != 1:
            raise DistributionTreeConflict(
                _("More than one distribution tree cannot be added to a " "repository version.")
            )


class RpmPublication(Publication, AutoAddObjPermsMixin):
    """
    Publication for "rpm" content.
    """

    TYPE = "rpm"
    checksum_type = models.TextField(choices=CHECKSUM_CHOICES)
    compression_type = models.TextField(null=True, choices=COMPRESSION_CHOICES)
    metadata_checksum_type = models.TextField(choices=CHECKSUM_CHOICES)
    package_checksum_type = models.TextField(choices=CHECKSUM_CHOICES)
    repo_config = models.JSONField(default=dict)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [
            ("manage_roles_rpmpublication", "Can manage roles on an RPM publication"),
        ]


class RpmDistribution(Distribution, AutoAddObjPermsMixin):
    """
    Distribution for "rpm" content.
    """

    TYPE = "rpm"
    SERVE_FROM_PUBLICATION = True
    repository_config_file_name = "config.repo"
    INVALID_REPO_ID_CHARS = r"[^\w\-_.:]"

    generate_repo_config = models.BooleanField(default=False)

    def content_handler(self, path):
        """Serve config.repo and repomd.xml.key."""
        if self.generate_repo_config and path == self.repository_config_file_name:
            repository, publication = self.get_repository_and_publication()
            if not publication:
                return
            if settings.DOMAIN_ENABLED:
                base_url = "{}/".format(
                    urlpath_sanitize(
                        settings.CONTENT_ORIGIN,
                        settings.CONTENT_PATH_PREFIX,
                        self.pulp_domain.name,
                        self.base_path,
                    )
                )
            else:
                base_url = "{}/".format(
                    urlpath_sanitize(
                        settings.CONTENT_ORIGIN,
                        settings.CONTENT_PATH_PREFIX,
                        self.base_path,
                    )
                )
            repo_config = publication.repo_config
            repo_config.pop("name", None)
            repo_config.pop("baseurl", None)
            val = textwrap.dedent(
                f"""\
                [{re.sub(self.INVALID_REPO_ID_CHARS, "", self.name)}]
                name={self.name}
                baseurl={base_url}
                """
            )
            for k, v in repo_config.items():
                val += f"{k}={v}\n"

            if "repo_gpgcheck" not in repo_config:
                val += "repo_gpgcheck=0\n"

            if "gpgcheck" not in repo_config:
                val += "gpgcheck=0\n"

            if "enabled" not in repo_config:
                val += "enabled=1\n"

            signing_service = repository.metadata_signing_service
            if signing_service:
                gpgkey_path = urlpath_sanitize(
                    base_url,
                    "/repodata/repomd.xml.key",
                )
                val += f"gpgkey={gpgkey_path}\n"

            return Response(body=val)

    def content_headers_for(self, path):
        """Return per-file http-headers."""
        headers = super().content_headers_for(path)
        base = os.path.basename(path)  # path.strip("/").split("/")[-1]
        if base in settings.NOCACHE_LIST:
            headers.update({"Cache-Control": "no-cache"})
        return headers

    def content_handler_list_directory(self, rel_path):
        """Return the extra dir entries."""
        retval = set()
        if self.generate_repo_config and rel_path == "":
            retval.add(self.repository_config_file_name)
        return retval

    def get_repository_and_publication(self):
        """Retrieves the repository and publication associated with this distribution if exists."""
        repository = publication = None
        if self.publication:
            publication = self.publication.cast()
            repository = publication.repository.cast()
        elif self.repository:
            repository = self.repository.cast()
            versions = repository.versions.all()
            publications = Publication.objects.filter(
                repository_version__in=versions, complete=True
            )
            try:
                publication = (
                    publications.select_related("repository_version")
                    .latest("repository_version", "pulp_created")
                    .cast()
                )
            except ObjectDoesNotExist:
                pass
        return repository, publication

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        permissions = [
            ("manage_roles_rpmdistribution", "Can manage roles on an RPM distribution"),
        ]
